float getGTRMonodirectionalShadowing0(float cotThetaV) {
	return 2.f/(1.f+sqrtf(1.f+1.f/(cotThetaV*cotThetaV)));
}

float getGTRMonodirectionalShadowing1(float sharpness, float cotThetaV) {
	float cotThetaV2=sqr(cotThetaV);
	float sharpness2=sqr(sharpness);
	float a=sqrtf(cotThetaV2+sharpness2);
	float b=sqrtf(cotThetaV2+1.0f);
	return cotThetaV*log(sharpness2)/(a-b+cotThetaV*log(sharpness2*(cotThetaV+b)/(cotThetaV+a)));
}

float getGTRMonodirectionalShadowing2(float sharpness, float cotThetaV) {
	return 2.f/(1.f+sqrtf(1.f+sqr(sharpness/cotThetaV)));
}

float getGTRMonodirectionalShadowing3(float sharpness, float cotThetaV) {
	float cotThetaV2=sqr(cotThetaV);
	float sharpness2=sqr(sharpness);
	float a=sqrtf(cotThetaV2+sharpness2);
	float b=sharpness2+1.0f;
	return 4.0f*cotThetaV*a*b/(2.0f*cotThetaV*b*(cotThetaV+a)+sharpness2*(3.0f*sharpness2+1.0f));
}

float getGTRMonodirectionalShadowing4(float sharpness, float cotThetaV) {
	float cotThetaV2=cotThetaV*cotThetaV;
	float sharpness2=sharpness*sharpness;
	float sharpness4=sharpness2*sharpness2;
	float a=8.f*(sharpness4+sharpness2+1.f);
	float b=sqrt(cotThetaV2+sharpness2);
	float b3=b*(cotThetaV2+sharpness2);
	return 2.f*cotThetaV*a*b3/
		(a*cotThetaV*(b3+cotThetaV*cotThetaV2)+3.f*sharpness2*(4.f*cotThetaV2*(2.f*sharpness4+sharpness2+1.f)+sharpness2*(5.f*sharpness4+2.f*sharpness2+1.f)));
}

/// Computes the monodirectional shadowing based on incoming/outgoing direction, micronormal, macronormal, glossiness and tail fall off.
/// If gtrGamma is any of the values 0, 1, 2, 3 or 4 there is an analytical solution of the double integral used to define the shadowing function
/// (see Microfacet Models for Refraction through Rough Surfaces EGSR 2007, Appendinx A and our document with details on deriving G1). 
/// If gtrGamma is not an integer and inside (0,4), a Cubic spline is constructed and used to get an interpolated value.
/// @param dir The direction of the incoming/outgoing ray.
/// @param microNormal The micronormal at the point of interest.
/// @param normal The macronormal at the point of interest.
/// @param sharpness The glossiness of the BRDF, between 0 and 1.
/// @param gtrGamma The gamma value (tail fall off).
float getGGXMonodirectionalShadowing(const Vector &dir, const Vector &microNormal, const Vector &normal, float sharpness, float gtrGamma) {
	float cosThetaV = dotf(dir, normal);

	if(cosThetaV<=1e-3f)
		return 0.0f;

	if(dotf(dir, microNormal) / cosThetaV <= 0.f)
		return 0.f;

	// when direction is collinear to the normal there is no shadowing
	if(cosThetaV>=1.0f-1e-6f)
		return 1.0f;

	float cotThetaV = cosThetaV / sqrtf(1.f - sqr(cosThetaV));

	float res=0.0f;

	// when gamma is any of the integer values 0, 1, 2, 3, 4 apply analytical solution
	if (gtrGamma<=0.01f)
		res=getGTRMonodirectionalShadowing0(cotThetaV);
	else if (fabsf(gtrGamma-1.0f)<=1e-2f)
		res=getGTRMonodirectionalShadowing1(sharpness, cotThetaV);
	else if (fabsf(gtrGamma-2.0f)<=1e-2f)
		res=getGTRMonodirectionalShadowing2(sharpness, cotThetaV);
	else if (fabsf(gtrGamma-3.0f)<=1e-2f)
		res=getGTRMonodirectionalShadowing3(sharpness, cotThetaV);
	else if(gtrGamma>=4.0f-1e-2f)
		res=getGTRMonodirectionalShadowing4(sharpness, cotThetaV);
	else {
		//gamma is not an integer. interpolate
		float G0=getGTRMonodirectionalShadowing0(cotThetaV);
		float G1=getGTRMonodirectionalShadowing1(sharpness, cotThetaV);
		float G2=getGTRMonodirectionalShadowing2(sharpness, cotThetaV);
		float G3=getGTRMonodirectionalShadowing3(sharpness, cotThetaV);
		float G4=getGTRMonodirectionalShadowing4(sharpness, cotThetaV);

		CubicSpline<5> spline;
		spline.addKnot(0.0f, G0);
		spline.addKnot(1.0f, G1);
		spline.addKnot(2.0f, G2);
		spline.addKnot(3.0f, G3);
		spline.addKnot(4.0f, G4);
		spline.construct(boundaryType_Natural);

		// clamp to avoid extrapolation
		res=spline.eval(clamp(gtrGamma, spline.knots[0].x, spline.knots[spline.numKnots-1].x));
	}

	return res;
}
