//*********************************************************
//         Author: Asen Atanasov
//           Year: Nov. 2012
//    Description: Implementation of cubic spline with 
//                 Complete, Natural, Periodic and NotAKnot
//                 boundaty conditions
//*********************************************************

#ifndef _SPLINE_H_
#define _SPLINE_H_

#include "solvetridiagonal.h"

/// Implementation of cubic spline interpolation with the most widely used boundary conditions.

/// Properties:
/// - Peacewise Cubic Polynomial - the spline is a cubic polynomial in every subinterval, defined by the input knots
/// - Interpolation              - passes through all of the knots
/// - Smoothness                 - continuous first and second derivatives (C2 smooth), everywhere in the interpolation interval
/// - Non-Locality               - the change of one knot, modifies the spline in the whole interval
/// - Energy Minimization        - among all the curves to satisfy the interpolation, smoothness and boundary condition constraints,
///                                the cubic spline is the curve with minimal energy, or minimal curvature.
///                                Formally, the spline is the curve with the minimal integral of the squared second derivative. 

/// During spline construction a linear system has to be solved. If only the interpolation and smoothness constraints are imposed,
/// then the system has two unknowns more than equations. Two get a well defined problem, two more equations must be provided.
/// This is done using different boundary conditions.

/// Boundary conditions:
/// - Complete - additionally the first derivatives of the spline at the endpoints must be provided. If a particular function is interpolated,
///              and these derivatives (or suitable approximations) are known, then this is probably the best choice.
/// - Natural  - the second derivatives of the spline at the endpoints are set to zero. If the first derivatives are not known, or we are
///              interpolating a table of values, then this is usually the choice for boundary conditions. The natural spline is also the
///              curve with minimal energy, among all splines corresponding to other boundary conditions. It well models an elastic rod,
///              which is forced to pass through the interpolation knots. The more bended the rod, the more potential energy it has.
/// - Periodic - it is imposed that the first and the second spline derivatives at the endpoints are equal. This means that the periodic
///              continuation of the spline (the spline is tiled infinitely many times) is also C2 smooth. It is additionally supposed that
///              the first and the last knot have the same ordinate value. It is used for interpolation on closed curves, e.g. a circle.
/// - NotAKnot - it is imposed that the spline has continuous third derivatives at the second and the penultimate interpolation knots.
///              In this way the same polynomials are defined in the first two subintervals and the last two respectively. Hence, the
///              second and the penultimate knots can be thought of as not real knots.

/// Usage:
/// Provide the interpolation knots either through addKnot(s) methods. Call construct() with the desired boundary conditions and,
/// for the case of Complete spline, the first derivatives da and db at the two end knots.
/// The knots either must be provided with increasing x coordinate or the sortKnots flag must be set to true.
/// Then the spline can be evaluated everywhere, along with its up to three derivatives. 
/// One can also directly access the array of polynomials. The number of polynomials is always numKnots-1.
/// To erase the knots, set numKnots member to 0.

/// Requirements:
/// At least 2 knots must be provided for Complete and Natural splines 
/// At least 4 knots must be provided for Periodic and NotAKnot splines.
/// The first and the last knot of a Periodic spline must have equal ordinate.
/// For knots with equal abscissae the problem is not defined, and the interpolation will fail.
/// More knots than MAX_KNOTS must not be provided.
/// Note that all of these requirements must be satisfied, for the code to work properly.
/// No internal checks are made, because the interpolation must be as fast as possible.

struct CubicPoly {

	float a;
	float b;
	float c;
	float d;

	CubicPoly() : a(0.0f), b(0.0f), c(0.0f), d(0.0f) {
	}

	CubicPoly(float a, float b, float c, float d) : a(a), b(b), c(c), d(d) {
	}

	float eval(float x) const {
		return ((a*x+b)*x+c)*x+d;
	}

	float deval(float x) const {
		return (3.0f*a*x+2.0f*b)*x+c;
	}

	float ddeval(float x) const {
		return 6.0f*a*x+2.0f*b;
	}

	float dddeval() const {
		return 6.0f*a;
	}
};

enum BoundaryType {
	boundaryType_Complete,
	boundaryType_Natural,
	boundaryType_Periodic,
	boundaryType_NotAKnot,
};

struct Knot {
	float x;
	float y;

	Knot() : x(0.0f), y(0.0f) {}
	Knot(float x, float y) : x(x), y(y) {}

	bool operator<(const Knot &p) {
		if(x!=p.x) return x<p.x;
		return y<p.y;
	}
};

template <int MAX_KNOTS>
class CubicSpline {

	// Finds the subinterval index, to which the point x belongs.
	int findPoly(float x) const {
		
		int left=0;
		int right=numKnots-1;

		while(right-left>1) {

			int mid=(left+right)/2;

			if(knots[mid].x<x) left=mid;
			else right=mid;
		}

		return left;
	}

	// Evaluate the spline moments for the case of complete boundary conditions. da and db are the derivatives at the endpoints of the spline interval [a, b].
	void evalMomentsComplete(float *m, float *h, int n , float da = 0.0f, float db = 0.0f) {

		float c[MAX_KNOTS];

		m[0] = 2.0f*h[0];
		m[n-1] = 2.0f*h[n-2];

		c[0] = 6.0f*( (knots[1].y-knots[0].y)/h[0] - da );
		c[n-1] = 6.0f*( db - (knots[n-1].y-knots[n-2].y)/h[n-2] );

		for(int i=1; i<n-1; i++) {

			m[i] = 2.0f*( h[i-1] + h[i] );
			c[i] = 6.0f*( (knots[i+1].y-knots[i].y)/h[i] - (knots[i].y-knots[i-1].y)/h[i-1] );
		}

		solveTridiagonal(h-1, m, h, c, n);
	}

	// Evaluate the spline moments for the case of natural boundary conditions.
	void evalMomentsNatural(float *m, float *h, int n) {

		float c[MAX_KNOTS];

		for(int i=1; i<n-1; i++) {

			m[i] = 2.0f*( h[i-1] + h[i] );
			c[i-1] = 6.0f*( (knots[i+1].y-knots[i].y)/h[i] - (knots[i].y-knots[i-1].y)/h[i-1] );
		}

		solveTridiagonal(h, m+1, h+1, c, n-2);

		m[0] = 0.0f;
		m[n-1] = 0.0f;
	}

	// Evaluate the spline moments for the case of periodic boundary conditions.
	void evalMomentsPeriodic(float *m, float *h, int n) {

		float c[MAX_KNOTS]; 
		float m2[MAX_KNOTS];

		m[0] = m2[0] = 2.0f*( h[0] + h[n-2] );
		c[0] = 6.0f*( (knots[1].y-knots[0].y)/h[0] -(knots[n-1].y-knots[n-2].y)/h[n-2] );

		for(int i=1; i<n-1; i++) {

			m[i] = m2[i] = 2.0f*( h[i-1] + h[i] );
			c[i] = 6.0f*( (knots[i+1].y-knots[i].y)/h[i] - (knots[i].y-knots[i-1].y)/h[i-1] );
		}

		solveTridiagonal(h-1, m, h, c, n-2);

		c[0] = -h[n-2];
		c[n-3] = -h[n-3];
		for(int i=1; i<n-3; i++) c[i] = 0.0f;

		solveTridiagonal(h-1, m2, h, c, n-2);

		m[n-2] = (c[n-2]-h[n-2]*m[0]-h[n-3]*m[n-3]) / (m[n-2]+h[n-2]*m2[0]+h[n-3]*m2[n-3]);

		for(int i=0; i<n-2; i++) {
			m[i] += m[n-2]*m2[i];
		}

		m[n-1] = m[0];
	}

	// Evaluate the spline moments for the case of not-a-knot boundary conditions.
	void evalMomentsNotAKnot(float *m, float *h, int n) {
		
		float a[MAX_KNOTS];
		float b[MAX_KNOTS];
		float c[MAX_KNOTS];

		for(int i=1; i<n-1; i++) {

			a[i] = h[i];
			b[i] = h[i+1];
			m[i+1] = 2.0f*( h[i] + h[i+1] );
			c[i-1] = 6.0f*( (knots[i+1].y-knots[i].y)/h[i] - (knots[i].y-knots[i-1].y)/h[i-1] );
		}

		float t1 = h[0]/h[1];
		float t2 = h[n-2]/h[n-3];

		m[1] = 3.0f*h[0] + 2.0f*h[1] + t1*h[0];
		m[n-2] = 3.0f*h[n-2] + 2.0f*h[n-3] + t2*h[n-2];

		a[n-3] = h[n-3] - t2*h[n-2];
		b[0] = h[1] - t1*h[0];

		solveTridiagonal(a, m+1, b, c, n-2);

		m[0] = m[1]*(1.0f+t1) - m[2]*t1;
		m[n-1] = m[n-2]*(1.0f+t2) - m[n-3]*t2;
	}

	// Initialize the array of polynomials using the spline moments m and the differences delta x, provided through h.
	void constructPolynomials(float *m, float *h, int n) {
		for(int i=0; i<n-1; i++) {
			CubicPoly &p=polys[i];

			p.a = (m[i+1]-m[i])/(6.0f*h[i]);
			p.b = 0.5f*m[i];
			p.c = (knots[i+1].y-knots[i].y)/h[i] - h[i]*(2.0f*m[i]+m[i+1])/6.0f;
			p.d = knots[i].y;
		}
	}

public:

	CubicSpline() : numKnots(0) {}

	void addKnot(float x, float y) {
		knots[numKnots].x=x;
		knots[numKnots].y=y;
		numKnots++;
	}

	void addKnot(const Knot &knot) {
		knots[numKnots++]=knot;
	}

	/// Constructs the spline by initializing the array of polynomials "polys", based on the array of interpolation knots "knots".
	/// @param boundaryType the boundary conditions imposed on the spline. The value is one of Complete, Natural, Periodic or NotAKnot values.
	/// @param da the value of the first derivative at the beginning of the spline interval. Used only for Complete boundary conditions.
	/// @param db the value of the first derivative at the end of the spline interval. Used only for Complete boundary conditions.
	/// @param sortKnots true to sort the knots by their x coordinate and false otherwise. By default, the knots are not sorted, and the user MUST provide them sorted.
	/// @retval true if the spline was constructed and false otherwise. For more information check the explanation in the beginning.
	int construct(BoundaryType boundaryType, float da = 0.0f, float db = 0.0f, int sortKnots=false) {

		int n = numKnots;
	
		if (sortKnots) {
			// Sort the input knots by their x-coordinate.
			VUtils::quickSort(knots, numKnots);
		}

		// allocate array of spline moments
		float m[MAX_KNOTS];

		// allocate and initialize subinterval length delta x
		float h[MAX_KNOTS];
		for(int i=0; i<n-1; i++) {
			h[i] = knots[i+1].x - knots[i].x;
		}

		// Compute the spline moments according to the provided boundary conditions.
		switch(boundaryType) {
			case boundaryType_Complete:
					  evalMomentsComplete(m, h, n, da, db); break;
			case boundaryType_Natural:
					  evalMomentsNatural(m, h, n); break;
			case boundaryType_Periodic:
					  evalMomentsPeriodic(m, h, n); break;
			case boundaryType_NotAKnot:
					  evalMomentsNotAKnot(m, h, n); break;
		}

		// Initialize the spline polynomials, based on the spline moments.
		constructPolynomials(m, h, n);

		return true;
	}

	/// Returns the spline value at the point x.
	/// @param x the position of evaluation.
	/// @note x must belong to the spline interval.
	float eval(float x) const {
		int idx=findPoly(x);
		return polys[idx].eval(x-knots[idx].x);
	}

	/// Returns the spline first derivative at the point x.
	/// @param x the position of evaluation.
	/// @note x must belong to the spline interval.
	float deval(float x) const {
		int idx=findPoly(x);
		return polys[idx].deval(x-knots[idx].x);
	}

	/// Returns the spline second derivative at the point x.
	/// @param x the position of evaluation.
	/// @note x must belong to the spline interval.
	float ddeval(float x) const {
		int idx=findPoly(x);
		return polys[idx].ddeval(x-knots[idx].x);
	}

	/// Returns the spline third derivative at the point x.
	/// @param x the position of evaluation.
	/// @note x must belong to the spline interval.
	float dddeval(float x) const {
		int idx=findPoly(x);
		return polys[idx].dddeval();
	}

	/// Returns the interpolation interval.
	/// @param[out] a the beginning of the interval
	/// @param[out] b the end of the interval.
	void getInterval(float &a, float &b) {
		a=knots[0].x;
		b=knots[numKnots-1].x;
	}

	/// Intentionally left public, to be manipulated either directly, or through addKnot(s) and eval() methods.
	Knot knots[MAX_KNOTS]; ///> Array of knots, subject to spline interpolation.
	CubicPoly polys[MAX_KNOTS-1]; ///> Array of spline polynomials.
	int numKnots; ///> The number of interpolation knots. It must be less than or equal to MAX_KNOTS. To erase the points set it to 0.
};

#endif //_SPLINE_H_